/*==============================================================================
	Title:			Three-phase PFC converter reference design
	File Name:		main.c
	Date:			2019/5/21
	Copyright:		Headspring Inc.
	Target CPU:		TMS320F28377S (200MHz)
	Target Board:	HSP0179-B00-0002A
	Target Inverter:HSP0179-B00-0001A
================================================================================*/

/*** Include *******************************************************/
#include "hs_math.h"
#include "hs_io.h"
#include "HSP0179-B00-0002A.h"

/*** Define ********************************************************/
/* PWM setting */
#define CARRIERFREQ_KHZ (50.0F)		/* Switching frequency[kHz] */
#define CARRIERFREQ_HZ (50000.0F)	/* Switching frequency[Hz] */
#define INTERRUPT_MASK (1U)			/* Decimation number */
#define DEADTIME_NSEC (200.0F)		/* Deadtime[nsec] */

/* Timer setting */
#define TIMER0INTERVAL_US (1000UL)	/* Timer0 period[usec] */

/* AD setting */
#define VUV_GAIN (814.5F)			/* UV Line voltage AD conversion gain */
#define VUV_OFFSET (0.0F)			/* UV Line voltage AD conversion offset */
#define VWV_GAIN (814.5F)			/* WV Line voltage AD conversion gain */
#define VWV_OFFSET (0.0F)			/* WV Line voltage AD conversion offset */
#define IU_GAIN (50.0F)				/* U-phase current AD conversion gain */
#define IU_OFFSET (0.0F)			/* U-phase current AD conversion offset */
#define IW_GAIN (50.0F)				/* W-phase current AD conversion gain */
#define IW_OFFSET (0.0F)			/* W-phase current AD conversion offset */
#define VDC_GAIN (468.0F)			/* DC voltage AD conversion gain */
#define VDC_OFFSET (468.0F)			/* DC voltage AD conversion offset */

/* Digital I/O */
#define DO_PRECHARGE (0x0001UL)		/* Digital output for driving magnetic contactor */

/* Mathematical constant */
#define MATHPI (3.1415926F)			/*  */
#define MATHSQRT3 (1.732051F)		/* 3 */

/* Controller parameter */
#define VQKP (0.02F)				/* q-axis voltage (PLL) Kp */
#define VQTI (0.5F)					/* q-axis voltage (PLL) Ti[msec] */
#define IDKP (20.0F)				/* d-axis current Kp */
#define IDTI (10.0F)				/* d-axis current Ti[msec] */
#define IQKP (20.0F)				/* q-axis current Kp */
#define IQTI (10.0F)				/* q-axis current Ti[msec] */
#define VDCKP (0.06F)				/* DC voltage Kp */
#define VDCTI (75.0F)				/* DC voltage Ti[msec] */
#define IDLIMIT (12.0F)				/* d-axis current limit[A] */
#define MRATELIMIT (1.0F)			/* modulation rate limit */

/* H/W parameters */
#define PHASE_ADVANCE_CAPACITANCE (6.8F)	/* Capacitance of phase advance capacitor[uF] */


/*** Typedef *******************************************************/
/* System state */
typedef enum{
	State_Error = 0,				/* 0:Error */
	State_Stop,						/* 1:Stop */
	State_Wait,						/* 2:Wait */
	State_Run,						/* 3:Run */
	State_Precharge,				/* 4:Precharge */
	State_Init,						/* 5:Initialization */
	State_NUM
}SystemState;

/* Protection */
typedef enum{
	Err_IU = 0,						/* 0:Iu OC protection*/
	Err_IW,							/* 1:Iw OC protection */
	Err_VDC,						/* 2:Vdc OV protection */
	Err_FLTU,						/* 3:U-phase gate driver error */
	Err_FLTV,						/* 4:V-phase gate driver error */
	Err_FLTW,						/* 5:W-phase gate driver error */
	Err_WDT,						/* 6:Watch dog timer error */
	Err_TRIP,						/* 7:Trip state */
	Err_NUM
}ErrState;

/* DI signal */
typedef enum{
	DI_FLTU = 0,					/* 0:U-phase gate driver error signal */
	DI_FLTV,						/* 1:V-phase gate driver error signal */
	DI_FLTW,						/* 2:W-phase gate driver error signal */
	DI_NUM
}DISignal;

/*** Functions *****************************************************/
void Init(void);					/* Initialization function */
void PWMIntFunc(void);				/* PWM interrupt function */
void Timer0IntFunc(void);			/* Timer0 interrupt function */

void JudgeError(void);				/* Judge error condition */

void SetState(void);				/* State machine */
void Start(void);					/* Pre-processing before running */
void Stop(void);					/* Exit process at stop running */

void CalcVacRMS(void);				/* Calculate RMS voltage */
void CalcLine2Phase(void);			/* Calculate phase-voltage from line-voltage */
void CalcPhase(void);				/* PLL */
void CalibADOffset(void);			/* Calibrate AD offset */

void CtrlVdc(void);					/* DC voltage controller */
void CtrlIdq(void);					/* AC current controller */
void CompDeadTime(void);			/* Deadtime compensation  */
void CalcMrate(void);				/* Calculate modulation rate */

/*** Variables *****************************************************/
UINT32 DIData = 0L;					/* Digital input state */

/* Sensor value */
FLOAT32	Vuv = 0.0F;					/* UV Line voltage[V] */
FLOAT32	Vwv = 0.0F;					/* WV Line voltage[V] */
FLOAT32	Iu = 0.0F;					/* U-phase current[A] */
FLOAT32	Iw = 0.0F;					/* W-phase current[A] */
FLOAT32	Vdc = 0.0F;					/* DC voltage[V] */

/* Calculated value */
FLOAT32	Vu = 0.0F;					/* U-phase voltage[V] */
FLOAT32	Vv = 0.0F;					/* V-phase voltage[V] */
FLOAT32	Vw = 0.0F;					/* W-phase voltage[V] */
FLOAT32 Va = 0.0F;					/* -axis voltage[V] */
FLOAT32 Vb = 0.0F;					/* -axis voltage[V] */
FLOAT32 Vd = 0.0F;					/* d-axis voltage[V] */
FLOAT32 Vq = 0.0F;					/* q-axis voltage[V] */
FLOAT32 Ia = 0.0F;					/* -axis current[A] */
FLOAT32 Ib = 0.0F;					/* -axis current[A] */
FLOAT32 Id = 0.0F;					/* d-axis current[A] */
FLOAT32 Iq = 0.0F;					/* q-axis current[A] */

/* Sensor value applied LPF */
FLOAT32 VdLPF = 0.0F;
FLOAT32 VqLPF = 0.0F;
FLOAT32 VqLPF2 = 0.0F;
FLOAT32 VdcLPF = 0.0F;
FLOAT32 IuLPF = 0.0F;
FLOAT32 IvLPF = 0.0F;
FLOAT32 IwLPF = 0.0F;
FLOAT32 IdLPF = 0.0F;
FLOAT32 IqLPF = 0.0F;

/* LPF cut-off frequency[Hz] */
FLOAT32 VdFc = 1.0F;
FLOAT32 VqFc = 1.0F;
FLOAT32 VqFc2 = 500.0F;
FLOAT32 VdcFc = 1000.0F;
FLOAT32 IuFc = 2000.0F;
FLOAT32 IwFc = 2000.0F;
FLOAT32 IdFc = 1000.0F;
FLOAT32 IqFc = 1000.0F;

/* Coefficient for calculating LPF */
FLOAT32 VdLPFCoeff = 0.0F;
FLOAT32 VqLPFCoeff = 0.0F;
FLOAT32 VqLPF2Coeff = 0.0F;
FLOAT32 VdcLPFCoeff = 0.0F;
FLOAT32 IdLPFCoeff = 0.0F;
FLOAT32 IqLPFCoeff = 0.0F;
FLOAT32 IuLPFCoeff = 0.0F;
FLOAT32 IwLPFCoeff = 0.0F;

FLOAT32 VdcRev = 0.0F;				/* Inverse of DC voltage[1/V] */
FLOAT32 OffsetVuv = 0.0F;			/* UV Line offset voltage[V] */
FLOAT32 OffsetVwv = 0.0F;			/* WV Line offset voltage[V] */
FLOAT32 OffsetIu = 0.0F;			/* U-phase offset current[A] */
FLOAT32 OffsetIw = 0.0F;			/* W-phase offset current[A] */
FLOAT32 OffsetVuv0 = 0.0F;			/* Temporary variable for calculating OffsetVuv */
FLOAT32 OffsetVwv0 = 0.0F;			/* Temporary variable for calculating OffsetVwv */
FLOAT32 OffsetIu0 = 0.0F;			/* Temporary variable for calculating OffsetIu */
FLOAT32 OffsetIw0 = 0.0F;			/* Temporary variable for calculating OffsetIw */
UINT16 CntOffset = 0U;				/* Counter used at the offset-compensation */
UINT16 FlagCalibEnd = 0U;			/* 1:Calibration ended, 0:Not calibration ended */

/* Protection */
UINT16 ErrorState = 0U;				/* Error state (non-latch) */
UINT16 ErrorStateLatch = 0U;		/* Error state (latch) */
FLOAT32 ErrThrVdcHi_V = 810.0F;		/* DC over voltage error threshold[V] */
FLOAT32 ErrThrIac_Apk = 11.0;		/* AC over current error threshold[A]iinstantaneous current valuej */
UINT16 FlagWDT = 0U;				/* 0:WDT overflow not occured, 1:WDT overflow occured */


/* State machine */
SystemState State = State_Init;		/* Current System state */
SystemState PreState = State_Init;	/* Previous state */
UINT16 InitEnd = 0U;				/* 0:Initializing, 1:Initialization complete */
FLOAT32 StopThrVac_Vrms = 250.0F;	/* State change threshold[Vrms] (->Stop) */
UINT16 RunStartStopReq = 1U;		/* 0:Run->Wait, 1:Wait->Run */
UINT16 FlagEntryWait = 0U;			/* Flag indicating Wait state */
UINT16 TimerEntryWait = 0U;			/* Counter for measuring time from entry Wait */

/* Calculation RMS */
FLOAT32 Vuv_Vrms = 0.0F;			/* UV Line Voltage[Vrms] */
FLOAT32 Vuv_Vrms0 = 0.0F;			/* UV Line Voltage[Vrms] (previous cycle) */
FLOAT32 VuvSum_V = 0.0F;			/* Integration Line Voltage(U-V) between period [V] */
INT16 SignVuv = 0U;					/* Sign Vuv (1:positive, -1:negative) */
INT16 SignVuv0 = 0U;				/* Sign Vuv at pre proccess (1:positive, -1:negative) */
UINT16 PeriodCount = 0U;			/* counter for calculating UV Line Voltage period */

/* PLL calculation */
FLOAT32 VqRef = 0.0F;				/* Vq reference value */
FLOAT32 ErrVq = 0.0F;				/* Vq error[V] */
FLOAT32 VqKp = VQKP;				/* Proportional gain */
FLOAT32 VqKi = VQKP * INTERRUPT_MASK / (VQTI * 0.001F * CARRIERFREQ_HZ);	/* Integral gain */
FLOAT32 VqTi = VQTI;				/* Time constant [msec] */
FLOAT32 VqPElement  = 0.0F;			/* Proportional term */
FLOAT32 VqIElement  = 0.0F;			/* Integral term */
FLOAT32 VqPIElement = 0.0F;			/* (Proportional term) + (Integral term) */
FLOAT32 Phase_rad = 0.0F;			/* Voltage phase of three-phase AC power source */
FLOAT32 RefPhase_rad = 0.0F;		/* Reference phase[rad] */
FLOAT32 RefDeltaPhase_rad = 0.0F;	/* Reference phase increment value[rad] */

/* AC current controller */
FLOAT32 IdRef = 0.0F;				/* Id reference value */
FLOAT32 ErrId = 0.0F;				/* Id error[A] */
FLOAT32 IdKp = IDKP;				/* Proportional gain */
FLOAT32 IdKi = IDKP * INTERRUPT_MASK / (IDTI * 0.001F * CARRIERFREQ_HZ);	/* Integral gain */
FLOAT32 IdTi = IDTI;				/* Time constant [msec] */
FLOAT32 IdPElement  = 0.0F;			/* Proportional term */
FLOAT32 IdIElement  = 0.0F;			/* Integral term */
FLOAT32 IdPIElement = 0.0F;			/* (Proportional term) + (Integral term) */
FLOAT32 VdOut = 0.0F;				/* Controller output */

FLOAT32 IqRef = 0.0F;				/* Iq reference value */
FLOAT32 ErrIq = 0.0F;				/* Iq error[A] */
FLOAT32 IqKp = IQKP;				/* Proportional gain */
FLOAT32 IqKi = IQKP * INTERRUPT_MASK / (IQTI * 0.001F * CARRIERFREQ_HZ);	/* Integral gain */
FLOAT32 IqTi = IQTI;				/* Time constant [msec] */
FLOAT32 IqPElement  = 0.0F;			/* Proportional term */
FLOAT32 IqIElement  = 0.0F;			/* Integral term */
FLOAT32 IqPIElement = 0.0F;			/* (Proportional term) + (Integral term) */
FLOAT32 VqOut = 0.0F;				/* Controller output */

/* DC voltage controller */
FLOAT32 IdLimitHi = IDLIMIT;		/* Id upperlimit[A] */
FLOAT32 IdLimitLo = -IDLIMIT;		/* Id lowerlimit[A] */
FLOAT32 VdcRef = 0.0F;				/* Vdc reference value */
FLOAT32 ErrVdc = 0.0F;				/* Vdc error[V] */
FLOAT32 VdcKp = VDCKP;				/* Proportional gain */
FLOAT32 VdcKi = VDCKP * (TIMER0INTERVAL_US * 0.001F * 0.001F) / (VDCTI * 0.001F);	/* Integral gain */
FLOAT32 VdcTi = VDCTI;				/* Time constant [msec] */
FLOAT32 VdcPElement  = 0.0F;		/* Proportional term */
FLOAT32 VdcIElement  = 0.0F;		/* Integral term */
FLOAT32 VdcPIElement = 0.0F;		/* (Proportional term) + (Integral term) */

/* Soft-start (DC voltage controller) */
FLOAT32 ConstVdcRef= 750.0F;		/* Target DC voltage[V] */
UINT16 FlagVdcRefRamp = 0U;			/* Flag indicating soft-start */
FLOAT32 DeltaVdcRefRamp = 0.0F;		/* Increment value of DC voltage reference */

FLOAT32 VaRef = 0.0F;				/* Va reference value[V] */
FLOAT32 VbRef = 0.0F;				/* Vb reference value[V] */
FLOAT32 VuRef = 0.0F;				/* Vu reference value[V] */
FLOAT32 VvRef = 0.0F;				/* Vv reference value[V] */
FLOAT32 VwRef = 0.0F;				/* Vw reference value[V] */
FLOAT32 Vmax = 0.0F;				/* Maximum value of phase-voltage[V] */
FLOAT32 Vmin = 0.0F;				/* Minimum value of phase-voltage[V] */
FLOAT32 Vavr = 0.0F;				/* Average value of phase-voltage[V] */

/* Deadtime compensation */
FLOAT32 PhaseUDeadTimeComp = 0.0F;	/* Added value to U-phase modulation rate */
FLOAT32 PhaseVDeadTimeComp = 0.0F;	/* Added value to V-phase modulation rate */
FLOAT32 PhaseWDeadTimeComp = 0.0F;	/* Added value to W-phase modulation rate */
FLOAT32 KDeadTime = 0.03F;			/* Coefficient of deadtime compensation */
FLOAT32 LimitDeadTimeComp = CARRIERFREQ_HZ * DEADTIME_NSEC * 0.001F * 0.001F * 0.001F;	/* Limit of deadtime compensation */

/* Modulation rate */
FLOAT32 MuRateOut = 0.0F;			/* U-phase */
FLOAT32 MvRateOut = 0.0F;			/* V-phase */
FLOAT32 MwRateOut = 0.0F;			/* W-phase */
FLOAT32 MRateLimHi = MRATELIMIT;	/* Upperlimit of modulation rate */
FLOAT32 MRateLimLo = -MRATELIMIT;	/* Lowerlimit of modulation rate */

/* Mathematical constant */
const FLOAT32 MathPi = MATHPI;						/*  */
const FLOAT32 Math2Pi = MATHPI * 2.0F;				/* 2 */
const FLOAT32 MathRev2Pi = 1.0F / (MATHPI * 2.0F);	/* 1/(2) */
const FLOAT32 MathSqrt3 = MATHSQRT3;				/* 3 */
const FLOAT32 MathRevSqrt3 = 1.0F / MATHSQRT3;		/* 1/3 */

const FLOAT32 ControlPeriod = INTERRUPT_MASK / CARRIERFREQ_HZ;	/* PWM interrupt period [sec] */
const FLOAT32 Capacitance_F = PHASE_ADVANCE_CAPACITANCE * 0.001F * 0.001F;	/*  Capacitance of phase advance capacitor[F] */

/* UpperLower Limit of AC voltage period */
const FLOAT32 PeriodCountMax = CARRIERFREQ_HZ / (INTERRUPT_MASK * 30.0F);
const FLOAT32 PeriodCountMin = CARRIERFREQ_HZ / (INTERRUPT_MASK * 70.0F);

/*=======================================================*/
/**
 * @brief   Main function.
 * @return  None.
 */
void main(void)
{
	Init();		/* Initialize */

	/* Main Loop */
	while(1)
	{
		ClearWDT(0U);				/* Clear WDT counter */
		FlagWDT = ReadFlagWDT();	/* Check WDT overflow */

		DIData = InputPIO();		/* DI read */

		UpdateSCOPE();				/* Communication with HSDT-DP */
	}
	
}

/*=======================================================*/
/**
 * @brief   Initialization function.
 * @return  None.
 */
void Init(void){

	DisableGlobalInterrupt();						/* Disable Global Interrupt */

	InitializeHSLIB_TypeB();						/* Initialize CPU Setting */
	InitializeBoard();								/* Initialize GPIO Setting */
	InitSCOPE();									/* Initialize Scope Function */

	/* Initialize AD Setting */
	InitADCA();
	InitADCB();
	InitADCC();
	InitADCD();
	
	SetupSocADCA(0U,5U,2U,75U);						/* Set up AD0 SOC0(ADCA ADCINA2) Vuv */
	SetupSocADCA(1U,5U,4U,75U);						/* Set up AD1 SOC1(ADCA ADCINA4) Vwv */
	SetupSocADCB(0U,5U,2U,75U);						/* Set up AD2 SOC0(ADCB ADCINB2) Iu */
	SetupSocADCD(0U,5U,14U,75U);					/* Set up AD3 SOC0(ADCD ADCIN14) Iw */
	SetupSocADCC(0U,5U,2U,75U);						/* Set up AD4 SOC0(ADCC ADCINC2) Vdc */
	
	InitRangeOffsetADCA(2U,VUV_GAIN,VUV_OFFSET);	/* Set range(814.5) and offset(0.0) for ADCINA2 */
	InitRangeOffsetADCA(4U,VWV_GAIN,VWV_OFFSET);	/* Set range(814.5) and offset(0.0) for ADCINA4 */
	InitRangeOffsetADCB(2U,IU_GAIN,IU_OFFSET);		/* Set range(50.0) and offset(0.0) for ADCINB2 */
	InitRangeOffsetADCD(14U,IW_GAIN,IW_OFFSET);		/* Set range(50.0) and offset(0.0) for ADCIN14 */
	InitRangeOffsetADCA(14U,IW_GAIN,IW_OFFSET);		/* Set range(50.0) and offset(0.0) for ADCIN14 */
	InitRangeOffsetADCC(2U,VDC_GAIN,VDC_OFFSET);	/* Set range(468.0) and offset(468.0) for ADCINC2 */

	/* Comparator Setting */
	InitCMPSS3(2U, 31U, 21U, 63U);					/* Set up Comparator Subsystem for Iu */
	InitCMPSS4(2U, 31U, 21U, 63U);					/* Set up Comparator Subsystem for Iw */
	InitCMPSS6(4U, 31U, 21U, 63U);					/* Set up Comparator Subsystem for Vdc */

	SetupLoOutputCMPSS3(3U);
	SetupHiOutputCMPSS3(3U);
	SetupLoOutputCMPSS4(3U);
	SetupHiOutputCMPSS4(3U);
	SetupHiOutputCMPSS6(3U);

	SetLovalueCMPSS3(-ErrThrIac_Apk, 3.0F);			/* Set threshold for over current protection */
	SetHivalueCMPSS3(ErrThrIac_Apk, 3.0F);			/* Set threshold for over current protection */
	SetLovalueCMPSS4(-ErrThrIac_Apk, 3.0F);			/* Set threshold for over current protection */
	SetHivalueCMPSS4(ErrThrIac_Apk, 3.0F);			/* Set threshold for over current protection */
	SetHivalueCMPSS6(ErrThrVdcHi_V, 3.0F);			/* Set threshold for over voltage protection */

	/* Initialize WDT */
	InitWDT(0U, 1U);								/* Overflow time:13.1msec, Stop WDT forbidden */

	/* Timer0 Setting */
	InitTIMER0_IntTime(TIMER0INTERVAL_US);			/* Initialize Timer0 Function */
	InitInterruptFunctionTIMER0(&Timer0IntFunc);	/* Initialize Interrupt Function */
	StartCountTIMER0();								/* Start Timer0 Counter */

	/* PWM Setting */
	/* Switching frequency[Hz], P-side Deadtime[nsec], N-side Deadtime[nsec],
		Update duty ratio both the peaks and valleys, Positive logic, Output pin(GPIO0-7) */
	InitPWM1to3(CARRIERFREQ_HZ, DEADTIME_NSEC, DEADTIME_NSEC, 2U, 0x0000U, 0x555U);

	/* Trip-Zone Setting */
	InitEpwmXbarPWM();						/* Initialize ePWM X-BAR */
	SetupInputXbar1(67U);					/* TRIP1 */
	SetupInputXbar2(68U);					/* TRIP2 */
	SetupInputXbar3(73U);					/* TRIP3 */
	SetupEpwmXbarTRIP4PWM(0x4F0UL);			/* TRIP4 : CMPSS3-Hi/Lo, CMPSS4-Hi/Lo, CMPSS6-Hi */
	InitDCAHTripPWM1(0x000FU);				/* TRIP1,2,3,4 */
	InitDCAHTripPWM2(0x000FU);				/* TRIP1,2,3,4 */
	InitDCAHTripPWM3(0x000FU);				/* TRIP1,2,3,4 */
	InitDCTripPWM1(0x000FU, 0x0002U);		/* Trip Combination Input, DCAEVT1.force */
	InitDCTripPWM2(0x000FU, 0x0002U);		/* Trip Combination Input, DCAEVT1.force */
	InitDCTripPWM3(0x000FU, 0x0002U);		/* Trip Combination Input, DCAEVT1.force */

	EnableTripControlPWM1(0x0040U);			/* Set up DCAEVT1 to trip source */
	EnableTripControlPWM2(0x0040U);			/* Set up DCAEVT1 to trip source */
	EnableTripControlPWM3(0x0040U);			/* Set up DCAEVT1 to trip source */

	/* Clear CMPSSLowLatch/CMPSSHighLatch */
	ClearLatchLoCMPSS3();
	ClearLatchHiCMPSS3();
	ClearLatchLoCMPSS4();
	ClearLatchHiCMPSS4();
	ClearLatchHiCMPSS6();

	/* Clear trip flag */
	ClearTripFlagPWM1();
	ClearTripFlagPWM2();
	ClearTripFlagPWM3();

	SetDutyPWM1to3(0.0F, 0.0F, 0.0F);		/* Set modulation rate */
	InitAdsocaPWM1(1U, 1U);					/* Initialization of AD-conversion starting trigger */
	EnableAdsocaPWM1();						/* Enable AD trigger */

	wait1us(1000);							/* Wait 1msec */
	
	InitInterruptPWM1(1U, INTERRUPT_MASK);	/* Initialize PWM1 Function */
	InitInterruptFunctionPWM1(&PWMIntFunc);	/* Initialize Interrupt Function */
	EnableInterruptPWM1();					/* Enable PWM1 Interrupt */
	EnableInterruptTIMER0();				/* Enable Timer0 Interrupt */
	EnableGlobalInterrupt();				/* Enable Global Interrupt */

	InitEnd = 1U;							/* Flag for state machine (Init->Stop) */
}

/*=======================================================*/
/**
 * @brief   Judge error condition.
 * @return  None.
 */
void JudgeError(void){
	UINT16 Error = 0U;

	Error |= (GetStatusLolatchCMPSS3() || GetStatusHilatchCMPSS3())  << Err_IU;	/* 0:Iu OC protection*/
	Error |= (GetStatusLolatchCMPSS4() || GetStatusHilatchCMPSS4()) << Err_IW;	/* 1:Iw OC protection */
	Error |= GetStatusHilatchCMPSS6() << Err_VDC;								/* 2:Vdc OV protection */
	Error |= ((DIData >> DI_FLTU) & 0x0001U) << Err_FLTU;						/* 3:U-phase gate driver error */
	Error |= ((DIData >> DI_FLTV) & 0x0001U) << Err_FLTV;						/* 4:V-phase gate driver error */
	Error |= ((DIData >> DI_FLTW) & 0x0001U) << Err_FLTW;						/* 5:W-phase gate driver error */
	Error |= (FlagWDT & 0x0001U) << Err_WDT;									/* 6:Watch dog timer error */
	Error |= (GetTripFlagPWM1() & 0x0001U) << Err_TRIP;							/* 7:TRIP state */

	ErrorState = Error;
	if((State != State_Init) && (State != State_Error))
	{
		ErrorStateLatch |=  ErrorState;
	}
}

/*=======================================================*/
/**
 * @brief   State machine.
 * @return  None.
 */
void SetState(void){

	switch(State){
	case State_Error:			/* 0:Error */
		break;
	case State_Stop:			/* 1:Stop */
		if(ErrorStateLatch != 0U)		/* Event:Error detection */
		{
			/* Exit Stop */
			State = State_Error;		/* Destination:Error */

			/* Entry Error */
			RunStartStopReq = 0U;
		}
		else if(Vuv_Vrms > 280.0F)		/* Event:AC Power Source connected */
		{
			/* Exit Stop */
			State = State_Precharge;	/* Destination:Precharge */

			/* Entry Precharge */
			VqIElement  = 0.0F;			/* Initialize PLL */
			VqPIElement = 0.0F;			/* Initialize PLL */
		}
		else
		{
			/* None */
		}
		break;
	case State_Wait:			/* 2:Wait */
		if(ErrorStateLatch != 0U)		/* Event:Error detection */
		{
			/* Exit Wait */
			FlagEntryWait = 0U;
			TimerEntryWait = 0U;

			State = State_Error;		/* Destination:Error */

			/* Entry Error */
			RunStartStopReq = 0U;
		}
		else if((Vuv_Vrms < 250.0F) || ((Vuv_Vrms - Vuv_Vrms0) < -20.0F))	/* Event:AC Power Source shutdown */
		{
			/* Exit Wait */
			FlagEntryWait = 0U;
			TimerEntryWait = 0U;

			State = State_Stop;			/* Destination:Stop */

			/* Entry Stop */
			OutputPIO(0UL);				/* Set DO0 to Low */
		}
		else if((TimerEntryWait > 599U) && (RunStartStopReq != 0U))		/* Event:Run request */
		{
			/* Exit Wait */
			FlagEntryWait = 0U;
			TimerEntryWait = 0U;

			State = State_Run;			/* Destination:Run */

			/* Entry Run */
			Start();
		}
		else
		{
			/* None */
		}
		
		break;
	case State_Run:				/* 3:Run */
		if(ErrorStateLatch != 0)		/* Event:Error detection */
		{
			/* Exit Run */
			Stop();
			
			State = State_Error;		/* Destination:Error */
			
			/* Entry Error */
			RunStartStopReq = 0U;
		}
		else if((Vuv_Vrms < 250.0F) || ((Vuv_Vrms - Vuv_Vrms0) < -20.0F))		/* Event:AC Power Source shutdown */
		{
			/* Exit Run */
			Stop();

			State = State_Stop;			/* Destination:Stop */
			
			/* Entry Stop */
			OutputPIO(0UL);				/* Set DO0 to Low */
		}
		else if(RunStartStopReq == 0U)		/* Event:Stop Request */
		{
			/* Exit Run */
			Stop();

			State = State_Wait;			/* Destination:Wait */
			
			/* Entry Wait */
			FlagEntryWait = 1U;
		}
		else
		{
			/* None */
		}
		
		break;
	case State_Precharge:		/* 4:Precharge */
		if(ErrorStateLatch != 0)		/* Event:Error detection */
		{
			/* Exit Precharge */
			/* None */

			State = State_Error;		/* Destination:Error */
			
			/* Entry Error */
			RunStartStopReq = 0U;
		}
		else if((Vuv_Vrms < 250.0F) || ((Vuv_Vrms - Vuv_Vrms0) < -20.0F))		/* Event:AC Power Source shutdown */
		{
			/* Exit Precharge */
			/* None */

			State = State_Stop;			/* Destination:Stop */
			
			/* Entry Stop */
			OutputPIO(0UL);				/* Set DO0 to Low */
		}
		else if(VdcLPF > (Vuv_Vrms * 1.34F))	/* Event:Precharge is finished (0.134 = 2*0.95) */
		{
			/* Exit Precharge */
			/* None */

			State = State_Wait;			/* Destination:Wait */
			
			/* Entry Wait */
			FlagEntryWait = 1U;
		}
		else
		{
			/* None */
		}
		break;
	case State_Init:				/* 5:Init */
		if(FlagCalibEnd != 0U){				/* Event:Initialization complete */
			/* Exit Init */
			/* None */

			State = State_Stop;			/* Destination:Stop */
			
			/* Entry Stop */
			OutputPIO(0UL);				/* Set DO0 to Low */
		}
		break;
	default:
		State = State_Error;
		break;
	}
}

/*=======================================================*/
/**
 * @brief   Pre-processing before running.
 * @return  None.
 */
void Start(void)
{
	/* Initialization integral term */
	IdIElement = 0.0F;
	IdPIElement = 0.0F;
	IqIElement = 0.0F;
	IqPIElement = 0.0F;
	VdcIElement = 0.0F;
	VdcPIElement = 0.0F;

	/* Set soft-start parameter */
	VdcRef = Vdc;
	DeltaVdcRefRamp = (ConstVdcRef - VdcRef) * 0.002F;
	FlagVdcRefRamp = 1U;

	dq2ab(VdLPF, VqLPF, Phase_rad, &VaRef, &VbRef);		/*  Inverse Park transformation */
	ab2uvw(VaRef, VbRef, &VuRef, &VvRef, &VwRef);		/*  Inverse Clarke transformation */

	if(VdcLPF > 0.1F)
	{
		VdcRev = 1.0F / VdcLPF;
	}
	else
	{
		VdcRev = 10.0;
	}

	MuRateOut = hslimit(VuRef * 2.0F * VdcRev, MRateLimHi, MRateLimLo);
	MvRateOut = hslimit(VvRef * 2.0F * VdcRev, MRateLimHi, MRateLimLo);
	MwRateOut = hslimit(VwRef * 2.0F * VdcRev, MRateLimHi, MRateLimLo);
	SetDutyPWM1to3(MuRateOut, MvRateOut, MwRateOut);	/* Set modulation rate */
	StartPWM1to3(0x555UL);								/* Start PWM */
}

/*=======================================================*/
/**
 * @brief   Exit process at stop running.
 * @return  None.
 */
void Stop(void)
{
	/* reset soft-start parameter */
	FlagVdcRefRamp = 0U;

	StopPWM1to3(0x555UL);					/* Stop PWM */
}

/*=======================================================*/
/**
 * @brief   Calculate RMS voltage.
 * @return  None.
 */
#pragma CODE_SECTION(CalcVacRMS, "ram_code")
void CalcVacRMS(void)
{
	FLOAT32 temp;

	if(Vuv > 0.0F)		/* positive/negative determination */
	{
		SignVuv = 1;
	}
	else
	{
		SignVuv = -1;
	}

	/* Detect Vuv zero cross */
	if((1 == SignVuv)&&(-1 == SignVuv0)&&(PeriodCount > PeriodCountMin))
	{
		if(PeriodCount > 1U)
		{
			temp = 1.0F / ((FLOAT32)PeriodCount * ControlPeriod);		/* Calculate AC frequency [Hz] */
		}
		else
		{
			temp = 1.0F;
		}

		Vuv_Vrms0 = Vuv_Vrms;
		Vuv_Vrms = hssqrt(temp * VuvSum_V * ControlPeriod) * 100.0F;	/* Calculate RMS voltage */

		/* Adjust reference phase */
		RefDeltaPhase_rad = Math2Pi * ControlPeriod * temp;
		RefPhase_rad = 1.5F * MathPi;
		
		IqRef = hslimit(-Vuv_Vrms * Math2Pi * temp * Capacitance_F, -0.3F, -1.4F);		/* Capacitor current compensation */

		VuvSum_V = 0.0F;
		PeriodCount = 0U;
	}
	else if(PeriodCount > PeriodCountMax)
	{
		Vuv_Vrms0 = Vuv_Vrms;
		Vuv_Vrms = 0.0F;
		IqRef = 0.0F;
		VuvSum_V = 0.0F;
		PeriodCount = 0U;
	}
	else
	{
		/* None */
	}

	SignVuv0 = SignVuv;		/* Save previous sign of Vuv */

	PeriodCount++;
	VuvSum_V += Vuv * Vuv * 0.01F * 0.01F;
}

/*=======================================================*/
/**
 * @brief   Calculate phase-voltage from line-voltage.
 * @return  None.
 */
#pragma CODE_SECTION(CalcLine2Phase, "ram_code")
void CalcLine2Phase(void)
{
	FLOAT32 TempVuw = 0.0F;

	TempVuw = Vuv - Vwv;
	Vu = (Vuv + TempVuw) * 0.333333F;
	Vv = -(Vuv + Vwv) * 0.333333F;
	Vw = -(Vu + Vv);
}

/*=======================================================*/
/**
 * @brief   PLL(Phase Locked Loop).
 * @return  None.
 */
#pragma CODE_SECTION(CalcPhase, "ram_code")
void CalcPhase(void)
{
	RefPhase_rad += RefDeltaPhase_rad;

	/* Folding */
	if(RefPhase_rad > Math2Pi)
	{
		RefPhase_rad -= Math2Pi;
	}

	ErrVq = VqLPF2 - VqRef;			/* Error */
	VqPElement  = VqKp * ErrVq;		/* Proportional term  */
	VqIElement  = hslimit((VqKi * ErrVq + VqIElement), MathPi * 1.1F, -MathPi * 1.1F);		/* Integral term */

	/* Folding */
	if(VqIElement > MathPi)
	{
		VqIElement -= Math2Pi;
	}
	else if(VqIElement < -MathPi)
	{
		VqIElement += Math2Pi;
	}
	else
	{
		/* None */
	}

	VqPIElement = hslimit((VqPElement + VqIElement + RefPhase_rad), MathPi * 3.0F, -MathPi);

	Phase_rad = VqPIElement;		/* Voltage phase of three-phase AC power source */
}

/*=======================================================*/
/**
 * @brief   Calibrate AD offset.
 * @return  None.
 */
void CalibADOffset(void)
{
	CntOffset++;
	OffsetVuv0 += -(Vuv * 0.01);
	OffsetVwv0 += -(Vwv * 0.01);
	OffsetIu0 += -(Iu * 0.01);
	OffsetIw0 += -(Iw * 0.01);

	if(CntOffset > 100U)		/* 100 samples */
	{
		CntOffset = 0U;
		OffsetVuv = OffsetVuv0;
		OffsetVwv = OffsetVwv0;
		OffsetIu = OffsetIu0;
		OffsetIw = OffsetIw0;
		FlagCalibEnd = 1U;
	}
}
/*=======================================================*/
/**
 * @brief   DC voltage controller.
 * @return  None.
 */
void CtrlVdc(void)
{
	ErrVdc = VdcRef - VdcLPF;			/* Error */
	VdcPElement = ErrVdc * VdcKp;		/* Proportional term */
	VdcIElement = hslimit(ErrVdc * VdcKi + VdcIElement, IdLimitHi, IdLimitLo);		/* Integral term */
	VdcPIElement = hslimit(VdcPElement + VdcIElement, IdLimitHi, IdLimitLo);
	IdRef = VdcPIElement;		/* Id reference value */
	VdcIElement = hslimit(IdRef - VdcPElement, IdLimitHi, IdLimitLo);		/* Anti-windup */
}

/*=======================================================*/
/**
 * @brief   AC current controller.
 * @return  None.
 */
#pragma CODE_SECTION(CtrlIdq, "ram_code")
void CtrlIdq(void)
{
	/* d-axis current controller */
	ErrId = IdLPF - IdRef;		/* Error */
	IdPElement = ErrId * IdKp;		/* Proportional term */
	IdIElement = hslimit(ErrId * IdKi + IdIElement, Vdc, -Vdc);		/* Integral term */
	IdPIElement = hslimit(IdPElement + IdIElement + VdLPF, Vdc, -Vdc);
	VdOut = IdPIElement;		/* Controller output (d-axis voltage) */
	IdIElement = hslimit(VdOut - IdPElement - VdLPF, Vdc, -Vdc);		/* Anti-windup */

	/* q-axis current controller */
	ErrIq = IqLPF - IqRef;		/* Error */
	IqPElement = ErrIq * IqKp;		/* Proportional term */
	IqIElement = hslimit(ErrIq * IqKi + IqIElement, Vdc, -Vdc);		/* Integral term */
	IqPIElement = hslimit(IqPElement + IqIElement + VqLPF, Vdc, -Vdc);
	VqOut = IqPIElement;		/* Controller output (q-axis voltage) */
	IqIElement = hslimit(VqOut - IqPElement - VqLPF, Vdc, -Vdc);		/* Anti-windup */
}

/*=======================================================*/
/**
 * @brief   Deadtime compensation.
 * @return  None.
 */
#pragma CODE_SECTION(CompDeadTime, "ram_code")
void CompDeadTime(void)
{
	IvLPF = -(IuLPF + IwLPF);

	PhaseUDeadTimeComp = hslimit(KDeadTime * IuLPF, LimitDeadTimeComp, -LimitDeadTimeComp);
	PhaseVDeadTimeComp = hslimit(KDeadTime * IvLPF, LimitDeadTimeComp, -LimitDeadTimeComp);
	PhaseWDeadTimeComp = hslimit(KDeadTime * IwLPF, LimitDeadTimeComp, -LimitDeadTimeComp);
}

/*=======================================================*/
/**
 * @brief   Calculate modulation rate.
 * @return  None.
 */
#pragma CODE_SECTION(CalcMrate, "ram_code")
void CalcMrate(void)
{
	FLOAT32 Vuvw[3] = {};		/* Vuw[0]:U-phase, Vuw[1]:V-phase, Vuw[2]:W-phase */
	UINT16 Loop = 0U;			/* Loop counter */

	dq2ab(VdOut, VqOut, Phase_rad, &VaRef, &VbRef);			/*  Inverse Park transformation */
	ab2uvw(VaRef, VbRef, &Vuvw[0], &Vuvw[1], &Vuvw[2]);		/*  Inverse Clarke transformation */

	/* Third harmonic injection */
	/* Search maximum value */
	Vmax = Vuvw[0];
	for(Loop = 1U; Loop < 3U; Loop++)
	{
		if(Vmax < Vuvw[Loop])
		{
			Vmax = Vuvw[Loop];
		}
	}
	/* Search minimum value */
	Vmin = Vuvw[0];
	for(Loop = 1U; Loop < 3U; Loop++)
	{
		if(Vuvw[Loop] < Vmin)
		{
			Vmin = Vuvw[Loop];
		}
	}
	Vavr = (Vmax + Vmin) * 0.5F;		/* Calculate average value */

	VuRef = Vuvw[0] - Vavr;				/* Vu reference value */
	VvRef = Vuvw[1] - Vavr;				/* Vv reference value */
	VwRef = Vuvw[2] - Vavr;				/* Vw reference value */

	if(VdcLPF > 0.1F)
	{
		VdcRev = 1.0F / VdcLPF;
	}
	else
	{
		VdcRev = 10.0;
	}

	MuRateOut = hslimit((VuRef * 2.0F * VdcRev + PhaseUDeadTimeComp), MRateLimHi, MRateLimLo);
	MvRateOut = hslimit((VvRef * 2.0F * VdcRev + PhaseVDeadTimeComp), MRateLimHi, MRateLimLo);
	MwRateOut = hslimit((VwRef * 2.0F * VdcRev + PhaseWDeadTimeComp), MRateLimHi, MRateLimLo);
}

/*=======================================================*/
/**
 * @brief   PWM Interrupt Function.
 * @return  None.
 */
#pragma CODE_SECTION(PWMIntFunc, "ram_code")
#pragma	INTERRUPT(PWMIntFunc)
void PWMIntFunc(void)
{
	AcknowledgePieINT();					/* Clear the PIE interrupt acknowledge flag */

	Vuv = ReadADCA(0U) + OffsetVuv;			/* Acquire sensor value : Vuv */
	Vwv = ReadADCA(1U) + OffsetVwv;			/* Acquire sensor value : Vwv */
	Iu = ReadADCB(0U) + OffsetIu;			/* Acquire sensor value : Iu */
	Iw = ReadADCD(0U) + OffsetIw;			/* Acquire sensor value : Iw */
	Vdc = ReadADCC(0U);						/* Acquire sensor value : Vdc */

	CalcVacRMS();							/* Calculate RMS voltage */

	uw2ab(Iu, Iw, &Ia, &Ib);				/* Clarke transformation */
	ab2dq(Ia, Ib, Phase_rad, &Id, &Iq);		/* Park transformation */

	/* Apply LPF */
	IdLPF = (IdLPFCoeff * IdLPF) + ((1.0F - IdLPFCoeff) * Id);
	IqLPF = (IqLPFCoeff * IqLPF) + ((1.0F - IqLPFCoeff) * Iq);
	IuLPF = (IuLPFCoeff * IuLPF) + ((1.0F - IuLPFCoeff) * Iu);
	IwLPF = (IwLPFCoeff * IwLPF) + ((1.0F - IwLPFCoeff) * Iw);

	CalcLine2Phase();						/* Calculate phase-voltage from line-voltage */

	uvw2ab(Vu, Vv, Vw, &Va, &Vb);			/* Clarke transformation */
	ab2dq(Va, Vb, Phase_rad, &Vd, &Vq);		/* Park transformation */

	/* Apply LPF */
	VdLPF = (VdLPFCoeff * VdLPF) + ((1.0F - VdLPFCoeff) * Vd);
	VqLPF = (VqLPFCoeff * VqLPF) + ((1.0F - VqLPFCoeff) * Vq);
	VqLPF2 = (VqLPF2Coeff * VqLPF2) + ((1.0F - VqLPF2Coeff) * Vq);

	CalcPhase();						/* PLL */

	if(State == State_Run)
	{
		CtrlIdq();						/* AC current controller */
		CompDeadTime();					/* Deadtime compensation */
		CalcMrate();					/* Calculate modulation rate */

		SetDutyPWM1to3(MuRateOut, MvRateOut, MwRateOut);	/* Set modulation rate */
	}

	ClearInterruptFlagPWM1();			/* Clear PWM1 interrupt flag */
}

/*=======================================================*/
/**
 * @brief   Timer0 Interrupt Function.
 * @return  None.
 */
#pragma	INTERRUPT(Timer0IntFunc)
void Timer0IntFunc(void)
{
	
	HSI_IERtable IntEnable;		/* New interrupt enable tables */
	HSI_IERtable Backup;		/* Pre-configuration interrupt enable table */

	AcknowledgePieINT();		/* Clear the PIE interrupt acknowledge flag */

	/* Interrupt enable state setting (for multiple interrupts) */
	IntEnable.pieier[0]  = 0x0000U;
	IntEnable.pieier[1]  = 0x0000U;
	IntEnable.pieier[2]  = 0x0001U;		/* Enable PWM1 interrupt */
	IntEnable.pieier[3]  = 0x0000U;
	IntEnable.pieier[4]  = 0x0000U;
	IntEnable.pieier[5]  = 0x0000U;
	IntEnable.pieier[6]  = 0x0000U;
	IntEnable.pieier[7]  = 0x0000U;
	IntEnable.pieier[8]  = 0x0000U;
	IntEnable.pieier[9]  = 0x0000U;
	IntEnable.pieier[10] = 0x0000U;
	IntEnable.pieier[11] = 0x0000U;
	IntEnable.ier = 0x4000U;			/* Enable Debugging communication interrupt */

	UpdateIerINT(&IntEnable, &Backup);	/* Update interrupt enable state (for multiple interrupts) */
	EnableGlobalInterrupt();			/* Enable Global Interrupt */

	/* 1msec measurement */
	if(FlagEntryWait == 1U)
	{
		if(TimerEntryWait > 599U)
		{
			FlagEntryWait = 0U;
		}
		else
		{
			TimerEntryWait++;
		}

		/* Delay turning on relay by 0.5sec */
		if(TimerEntryWait == 499U)		
		{
			OutputPIO(DO_PRECHARGE);		/* Set DO0 to High */
		}
	}

	/* VdcRef for soft-start */
	if(FlagVdcRefRamp == 1U)
	{
		if(hsfabs(VdcRef - ConstVdcRef) < 1.0F)
		{
			FlagVdcRefRamp = 0U;
			VdcRef = ConstVdcRef;
		}
		else
		{
			VdcRef += DeltaVdcRefRamp;		/* 500msec */
		}
	}

	JudgeError();						/* Judge error condition */
	SetState();							/* State machine */

	/* Calculate LPF parameter */
	VdLPFCoeff = 1.0F / (1.0F + ((2.0F * MATHPI * VdFc * INTERRUPT_MASK) / CARRIERFREQ_HZ));
	VqLPFCoeff = 1.0F / (1.0F + ((2.0F * MATHPI * VqFc * INTERRUPT_MASK) / CARRIERFREQ_HZ));
	VqLPF2Coeff = 1.0F / (1.0F + ((2.0F * MATHPI * VqFc2 * INTERRUPT_MASK) / CARRIERFREQ_HZ));
	IdLPFCoeff = 1.0F / (1.0F + ((2.0F * MATHPI * IdFc * INTERRUPT_MASK) / CARRIERFREQ_HZ));
	IqLPFCoeff = 1.0F / (1.0F + ((2.0F * MATHPI * IqFc * INTERRUPT_MASK) / CARRIERFREQ_HZ));
	IuLPFCoeff = 1.0F / (1.0F + ((2.0F * MATHPI * IuFc * INTERRUPT_MASK) / CARRIERFREQ_HZ));
	IwLPFCoeff = 1.0F / (1.0F + ((2.0F * MATHPI * IwFc * INTERRUPT_MASK) / CARRIERFREQ_HZ));
	VdcLPFCoeff = 1.0F / (1.0F + (2.0F * MATHPI * VdcFc * TIMER0INTERVAL_US * 0.001F * 0.001F));

	/* Apply LPF */
	VdcLPF = (VdcLPFCoeff * VdcLPF) + ((1.0F - VdcLPFCoeff) * Vdc);

	/* Calculate integral coefficient */
	VqKi = (VqKp * INTERRUPT_MASK) / (VqTi * 0.001F * CARRIERFREQ_HZ);
	IdKi = (IdKp * INTERRUPT_MASK) / (IdTi * 0.001F * CARRIERFREQ_HZ);
	IqKi = (IqKp * INTERRUPT_MASK) / (IqTi * 0.001F * CARRIERFREQ_HZ);
	VdcKi = (VdcKp * (TIMER0INTERVAL_US * 0.001F * 0.001F)) / (VdcTi * 0.001F);

	if(State == State_Run)
	{
		CtrlVdc();						/* DC voltage controller */
	}

	if(FlagCalibEnd == 0U)		/* Check calibration ended */
	{
		CalibADOffset();
	}

	DisableGlobalInterrupt();			/* Disable Global Interrupt */
	RestoreIerINT(&Backup);				/* Restore interrupt enabled state */

	ClearInterruptFlagTIMER0();			/* Clear Timer0 interrupt flag */
}
